#include <hpgcc49.h>
#include <hpgraphics.h>

#define WINDOW_WIDTH  				131
#define WINDOW_HEIGHT  				80
#define NULL 						0
#define PI         					((float) 3.141592654)
#define DEG_TO_RAD(ang) 			((ang) * PI / 180.0)
#define EPSILON_E5 					(float) (1E-5)
#define EPSILON_E3 					(float) (1E-3)

typedef struct {
	float x,y,z;
} VECTOR3D, POINT3D, *VECTOR3D_PTR, *POINT3D_PTR;

typedef struct {
	POINT3D p0; 									
	VECTOR3D n; 									
} PLANE, *PLANE_PTR;

typedef struct {
	union {
		float M[4][4];
		struct {
			float M00, M01, M02, M03;
			float M10, M11, M12, M13;
			float M20, M21, M22, M23;
			float M30, M31, M32, M33;
		};
	};
} MATRICE, *MATRICE_PTR;

typedef struct {
	union {
		float M[4]; 
		struct { float x,y,z,w;};
	};
} VECTOR4D, POINT4D, *VECTOR4D_PTR, *POINT4D_PTR;

typedef struct {
	union {     
		struct {float x,y,z,w;};    
		struct  {POINT4D  v;};             
	}; 
} VERTEX, *VERTEX_PTR;

typedef struct {					 
	int backface;
	int side;
	int vert[10];
} POLYGONE, *POLYGONE_PTR;

typedef struct {  
	char name[100];
	int faces[8];
	POINT4D world_pos;    
	VERTEX *vlist;
	POLYGONE *plist;  
	POLYGONE **pplist;   
} OBJECT, *OBJECT_PTR;

typedef struct {   					
	POINT4D pos;    								
	VECTOR4D dir;  	 							
	float view_dist;  								
	float fov;        							
	float near_z;    							
	float far_z;      					
	PLANE rt_plane;  					
	PLANE lt_plane;  						
	PLANE tp_plane;  						
	PLANE bt_plane;  						              
	float viewplane_width;     					
	float viewplane_height;  					  	
	float viewport_width; 						   
	float viewport_height;
	float viewport_center_x;  						
	float viewport_center_y;
	float aspect_ratio;
	MATRICE mcam;   								
	MATRICE mper; 							    
	MATRICE mscr; 							    
} CAMERA, *CAMERA_PTR;


//																								constants, macros, prototypes
const MATRICE ID_MAT = {1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1};

#define MAT_IDENTITY(m) {memcpy((void *)(m), (void *)&ID_MAT, sizeof(MATRICE));}
#define MAT_COPY(src_mat, dest_mat) {memcpy((void *)(dest_mat), (void *)(src_mat), sizeof(MATRICE) ); }
#define SWAP(a,b,t) {t=a; a=b; b=t;}

int compare_polygone(const void *arg1, const void *arg2);
void init_plane(PLANE_PTR plane, POINT3D_PTR p0, VECTOR3D_PTR normal);
void init_matrice(MATRICE_PTR ma, 
				float m00, float m01, float m02, float m03, float m10, float m11, float m12, float m13,
				float m20, float m21, float m22, float m23, float m30, float m31, float m32, float m33);				
void multiply_matrice(MATRICE_PTR ma, MATRICE_PTR mb, MATRICE_PTR mprod);
void build_rotation_matrice(float theta_x, float theta_y, float theta_z, MATRICE_PTR mrot);
void init_camera(CAMERA_PTR cam, POINT4D_PTR cam_pos, VECTOR4D_PTR cam_dir, float near_z, float far_z, 
					float fov, float viewport_width,  float viewport_height);
					
CAMERA cam; POINT4D cam_pos = { 0, 0, 0, 1 }; VECTOR4D cam_dir = { 0, 0, 0, 1 };

OBJECT surface; 
int raster_x[10], raster_y[10];
MATRICE mrot; float x_angle = 0.0; float y_angle = 0.0; float z_angle = 0.0;
int polygone, vertex; POINT4D presult; int col, row; float sum;
int num_verts, num_polys;
VECTOR3D *vertex_array;
float scale = 85.0;


int color[] = {HPG_COLOR_GRAY_1, HPG_COLOR_GRAY_2, HPG_COLOR_GRAY_3, HPG_COLOR_GRAY_4, HPG_COLOR_GRAY_5, HPG_COLOR_GRAY_6, HPG_COLOR_GRAY_7, 
				HPG_COLOR_GRAY_8, HPG_COLOR_GRAY_9, HPG_COLOR_GRAY_10, HPG_COLOR_GRAY_11, HPG_COLOR_GRAY_12, HPG_COLOR_GRAY_13, HPG_COLOR_GRAY_14};
				
int default_color[] = {2, 3, 5, 6, 9, 9, 12, 8}; 

#define MAX_POLYHEDRA		5+13+92

int offset_array[] = {13, 110, 276, 428, 830, 1151, 1410, 1697, 2183, 2663, 3220, 4165, 4793, 5480, 6651, 7823, 9175,
 11492, 13024, 13151, 13304, 13516, 13784, 14116, 14559, 14744, 14969, 15242, 15496, 15804, 15944, 16140, 16356, 16618,
 16934, 17224, 17579, 18034, 18597, 19271, 19666, 20173, 20800, 21538, 21733, 22037, 22425, 22812, 23292, 23771, 24367,
 24961, 25662, 26108, 26553, 27126, 27836, 28545, 29370, 30194, 31125, 32055, 32536, 33157, 33926, 34810, 35800, 35990,
 36216, 36477, 36743, 37045, 37348, 37691, 38034, 38408, 38866, 39370, 39874, 40415, 40679, 40900, 41159, 41518, 42126,
 42848, 44176, 45650, 47124, 48735, 50100, 51471, 52842, 54210, 55439, 56679, 57919, 59157, 60252, 61347, 62445, 63397,
 63616, 64053, 64313, 64617, 64937, 65311, 65731, 66058};

void color_selector();
int read_polyhedra_data();

char buffer[80];
FILE *fpoly;
int flag_solid = 1;
int flag_backface = 1;
int tick_count = 3;
int current_side = 3;
short name_length;

int polyhedra_index = 0;
int  flag_name;

int main() {	
	unsigned char ticks = sys_setRTCTickPeriod(1);
	int i,j; hpg_clear(); sys_slowOff();
		
	if( (fpoly = fopen("POLYHEDRA.DAT", "rb")) == NULL) { printf("Error:couldn't open POLYHEDRA.DAT"); keyb_getkey(1); return (0);}
	if(!read_polyhedra_data()) return 0;
	
	init_camera(&cam, &cam_pos, &cam_dir, 50.0, 500.0, 90.0, WINDOW_WIDTH, WINDOW_HEIGHT);
	surface.world_pos.x = 0; surface.world_pos.y = 0; surface.world_pos.z = 200.0; surface.world_pos.w = 1.0;
	hpg_set_mode_gray16(1); hpg_clear(); hpg_flip(); hpg_clear();		
	
	while (!keyb_isON()) {	
		y_angle += 0.02;
		
		if (keyb_isAnyKeyPressed()) {		
			if(keyb_isUp()) x_angle += 0.1; else if(keyb_isDown()) x_angle += -0.1;
			else if(keyb_isLeft()) y_angle += 0.1; else if(keyb_isRight()) y_angle += -0.1;
			else if(keyb_isKeyPressed(4,0)) {z_angle += 0.1;} else if(keyb_isKeyPressed(3,0)) {z_angle -= 0.1;}
			
			else if(keyb_isKeyPressed(0,5)) {if(surface.world_pos.z > 100.0) surface.world_pos.z -= 5.0;}
			else if(keyb_isKeyPressed(0,4)) {surface.world_pos.z += 5.0;}
			
			else if (keyb_isLS()) { 
				if(keyb_isKeyPressed(6,4)) {tick_count = tick_count <= 0 ? 0 : (tick_count - 1); }
				
				else if(keyb_isKeyPressed(7,1)){polyhedra_index += 40;if(polyhedra_index > MAX_POLYHEDRA-1) polyhedra_index -= MAX_POLYHEDRA; read_polyhedra_data();}	
				else if(keyb_isKeyPressed(7,0)){polyhedra_index -= 40;if(polyhedra_index < 0) polyhedra_index += MAX_POLYHEDRA; read_polyhedra_data();}
			}
//																								freeze the screen			
			else if(keyb_isKeyPressed(5,0)) keyb_getkey(1);
			
			else if(keyb_isKeyPressed(5,6)) {flag_backface = 1; flag_solid = flag_solid == 1 ? 0 : 1; }		
			else if(keyb_isKeyPressed(6,6)) {flag_solid = 0;flag_backface = flag_backface == 1 ? 0: 1;}	
			
			else if(keyb_isKeyPressed(7,1)){polyhedra_index = polyhedra_index == MAX_POLYHEDRA-1 ? 0 : polyhedra_index + 1; read_polyhedra_data();}	
			else if(keyb_isKeyPressed(7,0)){polyhedra_index = polyhedra_index == 0 ? MAX_POLYHEDRA-1 : polyhedra_index - 1; read_polyhedra_data();}
			
			else if(keyb_isKeyPressed(6,4)) {tick_count += 1; }
			
			else if(keyb_isKeyPressed(5,2)) color_selector();
//																								display info
			else if (keyb_isKeyPressed(6,5)) {
				j = 2;
				hpg_clear(); hpg_draw_text(surface.name, 65 - (name_length<<1), 0);
				if(flag_name) {
					j = 3;
					hpg_draw_text(&surface.name[name_length+1], 65 - (strlen(&surface.name[name_length+1])<<1), 8);
				}
				for(i = 0; i < 8; i++) {
					if(surface.faces[i] != 0) { 
						sprintf(buffer, "%2d..%d_sided faces", surface.faces[i], i+3);	
						hpg_draw_text(buffer, 5, 8 * j++);
					}
				}
				hpg_flip(); keyb_getkey(1); 	
			}	
			sys_waitRTCTicks(3);
		}
		
		for( vertex = 0; vertex < num_verts; vertex++) {			
			surface.vlist[vertex].x = vertex_array[vertex].x * scale; surface.vlist[vertex].y = vertex_array[vertex].y * scale;
			surface.vlist[vertex].z = vertex_array[vertex].z * scale; surface.vlist[vertex].w = 1.0;
		}
	
		for (polygone = 0; polygone < num_polys; polygone++) surface.plist[polygone].backface = 0;
		build_rotation_matrice(x_angle, y_angle, z_angle, &mrot);	

		for (vertex = 0; vertex < num_verts; vertex++) {
			for (col = 0; col < 4; col++) {       						
				sum = 0; 
				for (row = 0; row < 4; row++) { sum += (surface.vlist[vertex].v.M[row] * mrot.M[row][col]);} 
				presult.M[col] = sum;
			} 
			surface.vlist[vertex].v.x = presult.x + surface.world_pos.x; surface.vlist[vertex].v.y = presult.y + surface.world_pos.y;
			surface.vlist[vertex].v.z = presult.z + surface.world_pos.z; surface.vlist[vertex].v.w = 1;	
			for (col = 0; col < 4; col++) {       						
				sum = 0; 
				for (row = 0; row < 4; row++) { sum += (surface.vlist[vertex].v.M[row] * cam.mcam.M[row][col]);} 
				presult.M[col] = sum;
			} 
			surface.vlist[vertex].v.x = presult.x; surface.vlist[vertex].v.y = presult.y;
			surface.vlist[vertex].v.z = presult.z; surface.vlist[vertex].v.w = presult.w;	
		}
//																								there is a issue with these 2 polyhedra: j26 & j37
		if(polyhedra_index != 43 && polyhedra_index != 54 ) {
			if( flag_backface) {
				for (polygone = 0; polygone < num_polys; polygone++) {
					POLYGONE_PTR curr_poly = &surface.plist[polygone];
					VECTOR4D u, v, n, view;
					u.x = surface.vlist[curr_poly->vert[1]].x - surface.vlist[curr_poly->vert[0]].x;		
					u.y = surface.vlist[curr_poly->vert[1]].y - surface.vlist[curr_poly->vert[0]].y;
					u.z = surface.vlist[curr_poly->vert[1]].z - surface.vlist[curr_poly->vert[0]].z;
					u.w = 1;
					v.x = surface.vlist[curr_poly->vert[2]].x - surface.vlist[curr_poly->vert[0]].x;		
					v.y = surface.vlist[curr_poly->vert[2]].y - surface.vlist[curr_poly->vert[0]].y;
					v.z = surface.vlist[curr_poly->vert[2]].z - surface.vlist[curr_poly->vert[0]].z;
					v.w = 1;
					n.x = ( (u.y * v.z) - (u.z * v.y) ); n.y = -( (u.x * v.z) - (u.z * v.x) );											
					n.z = ( (u.x * v.y) - (u.y * v.x) ); n.w = 1;
					view.x = cam.pos.x - surface.vlist[curr_poly->vert[0]].x;										
					view.y = cam.pos.y - surface.vlist[curr_poly->vert[0]].y;
					view.z = cam.pos.z - surface.vlist[curr_poly->vert[0]].z;
					view.w = 1;
					float dp = (n.x * view.x) + (n.y * view.y) + (n.z * view.z); 											
					if (dp <= EPSILON_E5) curr_poly->backface |= 1;
				}
			}
		}
		else {qsort((void *) surface.pplist, num_polys, sizeof(POLYGONE_PTR), compare_polygone);}
			
		float alpha = (0.5 * cam.viewport_width - 0.5); float beta = (0.5 * cam.viewport_height - 0.5);
		for (vertex = 0; vertex < num_verts; vertex++) {
			float z = surface.vlist[vertex].z; if(z == 0.0) z = EPSILON_E5;	
			surface.vlist[vertex].x = cam.view_dist * surface.vlist[vertex].x / z;
			surface.vlist[vertex].y = cam.view_dist * surface.vlist[vertex].y * cam.aspect_ratio / z;
			surface.vlist[vertex].x = alpha + alpha * surface.vlist[vertex].x;
			surface.vlist[vertex].y = beta - beta * surface.vlist[vertex].y;
		}
		
		hpg_clear();
		for (polygone = 0; polygone < num_polys; polygone++) {
			if(surface.plist[polygone].backface == 1) continue;
			for(i = 0; i < surface.pplist[polygone]->side; i++) {
				raster_x[i] = (int) surface.vlist[surface.pplist[polygone]->vert[i]].x;
				raster_y[i] = (int) surface.vlist[surface.pplist[polygone]->vert[i]].y;
			}
			if(flag_solid) {
				hpg_set_color(hpg_stdscreen, color[default_color[surface.pplist[polygone]->side - 3]]); 
				hpg_fill_polygon(raster_x, raster_y, surface.pplist[polygone]->side);	
			}
				hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK);hpg_draw_polygon(raster_x, raster_y, surface.pplist[polygone]->side);			
		}			
		sys_waitRTCTicks(tick_count);hpg_flip();
	}
	sys_slowOn(); sys_setRTCTickPeriod(ticks); fclose(fpoly); return 1;
}

int read_polyhedra_data(){
	int i; for(i = 0; i < 8; surface.faces[i++] = 0);
	fseek(fpoly, offset_array[polyhedra_index], SEEK_SET);	
//																								read the polyheddron name
	fread(&name_length, 1, sizeof(short), fpoly); 
	fread(&surface.name, 1, sizeof(char)*name_length, fpoly); surface.name[name_length] = '\0';
// 																								split the name in 2 if too long	
	flag_name = 0; char *p, *q;
	if(polyhedra_index > 17 && name_length > 32) {
		flag_name = 1; p = strrchr(surface.name,'('); *--p = '\0';
		q = strrchr(surface.name,' '); *q = '\0'; *p = ' ';
		name_length = strlen(surface.name);
	} 
//																								read the vertices
	fread(&num_verts, 1, sizeof(short), fpoly);
	if((vertex_array = (VECTOR3D *) malloc(num_verts * sizeof(VECTOR3D))) == NULL  ||
		(surface.vlist = (VERTEX *) malloc(num_verts * sizeof(VERTEX))) == NULL) {printf("Error allocation memory: vertices\n"); keyb_getkey(1); return 0;}
	for(vertex = 0; vertex < num_verts; vertex++) {fread(&vertex_array[vertex], 1, sizeof(VECTOR3D), fpoly);}
//																								read the faces
	fread(&num_polys, 1, sizeof(short), fpoly);
	if((surface.plist = (POLYGONE *) calloc(num_polys, sizeof(POLYGONE))) == NULL ) {printf("Error allocation memory: plist\n"); keyb_getkey(1); return 0;}	
	for(polygone = 0; polygone < num_polys; polygone++) { 
		fread(&surface.plist[polygone].side, 1, sizeof(short), fpoly);
		surface.faces[surface.plist[polygone].side - 3]++;
		for(i = 0; i < surface.plist[polygone].side; i++) {
			fread(&surface.plist[polygone].vert[i], 1, sizeof(short), fpoly);
		}
	}
	if((surface.pplist = (POLYGONE_PTR *) malloc((num_polys) * sizeof(POLYGONE_PTR))) == NULL ) {printf("Error allocation memory: pplist\n"); keyb_getkey(1); return 0;}
	for(i = 0; i < num_polys; i++) surface.pplist[i] = &surface.plist[i];
	return 1;
}

void color_selector() {
	int i; 
	while (1) {	
		hpg_clear(); hpg_draw_text("    FACES GRAY LEVEL SELECTOR", 0, 0);
		for(i = 3; i < 11; i++) {	
			hpg_set_color(hpg_stdscreen, color[default_color[i-3]]); hpg_fill_rect(56, 9*(i-2), 80, (9*(i-2)) + 6);	
			hpg_set_color(hpg_stdscreen, HPG_COLOR_BLACK);
			if(i == current_side) sprintf(buffer, "%2d-sided: -->", i); else sprintf(buffer, "%2d-sided:", i); 
			hpg_draw_text(buffer, 0, 9*(i-2));	
			sprintf(buffer, "gray_%d", default_color[i-3]+1); hpg_draw_text(buffer, 95, 9*(i-2));	
		}
		sys_waitRTCTicks(3);hpg_flip();
		if (keyb_isAnyKeyPressed()) {		
			if (keyb_isKeyPressed(4,1)) return;
			else if(keyb_isUp()) {current_side = current_side == 3 ? 10 : (current_side - 1);}
			else if(keyb_isDown()) {current_side = current_side == 10 ? 3 : (current_side + 1);}
			else if(keyb_isLeft()) {default_color[current_side-3] = default_color[current_side-3] == 0 ? 13 : (default_color[current_side-3] - 1);}
			else if(keyb_isRight()) {default_color[current_side-3] = default_color[current_side-3] == 13 ? 0 : (default_color[current_side-3] + 1);}
			sys_waitRTCTicks(3);
		}	
	}
}

//																								engine routine													
int compare_polygone(const void *arg1, const void *arg2) {
	int i;float z1 = 0.0; float z2 = 0.0; POLYGONE_PTR poly_1, poly_2;
	poly_1 = *((POLYGONE_PTR *)(arg1)); poly_2 = *((POLYGONE_PTR *)(arg2));
	for(i = 0; i < poly_1->side; i++) {z1 += surface.vlist[poly_1->vert[i]].z;} z1 /= poly_1->side;
	for(i = 0; i < poly_2->side; i++) {z2 += surface.vlist[poly_2->vert[i]].z;} z2 /= poly_2->side;
	if (z1 > z2) return(-1); else if (z1 < z2) return(1); else return(0);
} 

void init_plane(PLANE_PTR plane, POINT3D_PTR p0, VECTOR3D_PTR normal) {
	plane->p0.x = p0->x;plane->p0.y = p0->y;plane->p0.z = p0->z;								
	float length = (float) sqrt( normal->x * normal->x + normal->y * normal->y + normal->z * normal->z);
	if (length < EPSILON_E5) return;							
	plane->n.x = normal->x / length; plane->n.y = normal->y / length; plane->n.z = normal->z / length;
} 

void init_matrice(MATRICE_PTR ma,	float m00, float m01, float m02, float m03, float m10, float m11, float m12, float m13,
										float m20, float m21, float m22, float m23, float m30, float m31, float m32, float m33) {
	ma->M00 = m00; ma->M01 = m01; ma->M02 = m02; ma->M03 = m03;
	ma->M10 = m10; ma->M11 = m11; ma->M12 = m12; ma->M13 = m13;
	ma->M20 = m20; ma->M21 = m21; ma->M22 = m22; ma->M23 = m23;
	ma->M30 = m30; ma->M31 = m31; ma->M32 = m32; ma->M33 = m33;
} 

void multiply_matrice(MATRICE_PTR ma, MATRICE_PTR mb, MATRICE_PTR mprod) {
	int row,col,index;
	for (row=0; row<4; row++) {
		for (col=0; col<4; col++) {
			float sum = 0; 
			for (index = 0; index < 4; index++) {
				sum += (ma->M[row][index] * mb->M[index][col]);
			} 
			mprod->M[row][col] = sum;
        } 
    } 
} 

void build_rotation_matrice(float theta_x, float theta_y, float theta_z, MATRICE_PTR mrot)  {
	MATRICE mx, my, mz, mtmp;       							
	float sin_theta = 0, cos_theta = 0;						   	  							
	MAT_IDENTITY(mrot);										
	cos_theta = cos(theta_x);
	sin_theta = sin(theta_x);
	init_matrice(&mx, 	1,    0,         0,         0,
						0,    cos_theta, sin_theta, 0,
						0,   -sin_theta, cos_theta, 0,
						0,    0,         0,         1);
	cos_theta = cos(theta_y);
	sin_theta = sin(theta_y);
	init_matrice(&my,	cos_theta, 0, -sin_theta, 0,  
						0,         1,  0,         0,
						sin_theta, 0,  cos_theta,  0,
						0,         0,  0,          1);
	cos_theta = cos(theta_z);
	sin_theta = sin(theta_z);
	init_matrice(&mz, 	cos_theta, 	sin_theta, 0, 0,  
						-sin_theta, cos_theta, 0, 0,
						0,         	0,         1, 0,
						0,         	0,         0, 1);
	multiply_matrice(&mx, &my, &mtmp);
	multiply_matrice(&mtmp, &mz, mrot);
}

void init_camera(CAMERA_PTR cam, POINT4D_PTR cam_pos, VECTOR4D_PTR cam_dir, float near_z, float far_z, float fov, float viewport_width, float viewport_height) {
		cam->pos.x = cam_pos->x;cam->pos.y = cam_pos->y;cam->pos.z = cam_pos->z;cam->pos.w = cam_pos->w;
		cam->dir.x = cam_dir->x;cam->dir.y = cam_dir->y;cam->dir.z = cam_dir->z;cam->dir.w = cam_dir->w;
		cam->near_z = near_z; cam->far_z  = far_z;     	
		cam->viewport_width  = viewport_width; cam->viewport_height = viewport_height;
		cam->viewport_center_x = (viewport_width-1)/2; cam->viewport_center_y = (viewport_height-1)/2;
		cam->aspect_ratio = (float) viewport_width / (float) viewport_height;
		MAT_IDENTITY(&cam->mcam); MAT_IDENTITY(&cam->mper); MAT_IDENTITY(&cam->mscr);
		cam->fov = fov; cam->viewplane_width  = 2.0; cam->viewplane_height = 2.0/cam->aspect_ratio;
		float tan_fov_div2 = tan(DEG_TO_RAD(fov / 2));
		cam->view_dist = (0.5) * (cam->viewplane_width) * tan_fov_div2;
		POINT3D pt_origin; pt_origin.x = 0.0; pt_origin.y = 0.0; pt_origin.z = 0.0;
		VECTOR3D vn; 
		vn.x =  1.0;vn.y =  0.0;vn.z = -1.0; init_plane(&cam->rt_plane, &pt_origin, &vn);
		vn.x = -1.0;vn.y =  0.0;vn.z = -1.0; init_plane(&cam->lt_plane, &pt_origin, &vn);
		vn.x =  0.0;vn.y =  1.0;vn.z = -1.0; init_plane(&cam->tp_plane, &pt_origin, &vn);
		vn.x =  0.0;vn.y = -1.0;vn.z = -1.0; init_plane(&cam->bt_plane, &pt_origin, &vn); 		
}










